//   LZW 
//    ,    ( 9   )
// : 8      (512  65536 )
// : 
//
// icq:133575351  email:lawnmower-man@mail.ru

#include<stdio.h>
#include<memory.h>
#include<malloc.h>

#define LZ_END	0		//  
//[1..256] -   
#define LZ_START 257	//    

#define CAPABLITY 65536	//!<  

typedef  unsigned char  BYTE;
typedef  unsigned short CELL;	// 16  (   )


class LZW{
	struct dict_st{
		CELL prefix;	//!< 
		CELL   next;	//!<    
		CELL  child;	//!<   
		BYTE    sym;	//!<  
	};	//!<     (0        )
protected:
	int count;		//!<     
	int bits_count;	//!<     
	dict_st *table;	//!<  
	void Init();					//!<  
	void Update(int,BYTE);			//!<  
	int  Find(int,BYTE);			//!<    
public:
	LZW();
	~LZW();
};


class LZW_Encoder:LZW{
	BYTE byte,byte_shift; 
	void StartWrite() {byte=0,byte_shift=1;}
	void StopWrite(FILE* out) {fputc(byte,out);}
	void WriteCode(FILE*,int);		//!<    
public:
	void Pack(FILE*,FILE*);			//!< 
	static void _Encode(FILE*,FILE*);
};


class LZW_Decoder:LZW{
	BYTE byte,byte_shift; 
	char *stack;		//<!    
	int stack_size;		//<!    
	void StartRead() {byte=byte_shift=0;}
	int  ReadCode(FILE*);
	void WriteSym(FILE*,int);		//!<    
public:
	LZW_Decoder();
	~LZW_Decoder();
	void Unpack(FILE*,FILE*);		//!< 
	static void _Decode(FILE*,FILE*);
};



LZW::LZW()
{
	table=new dict_st[CAPABLITY];
}

LZW::~LZW()
{
	delete[]table;
}


void LZW::Init()
{
	memset(table,0,sizeof(dict_st)*CAPABLITY);
	for(int n=0; n<256; n++) {
		table[n+1].next=n+2;
		table[n+1].sym=n;
	}
	table[n].sym=255;
	count=LZ_START;
	bits_count=9;
}

//! \param code  
//! \param sym   
void LZW::Update(int code,BYTE sym)
{
	table[count].prefix=code;
	table[count].next=table[code].child;
	table[code ].child=count;
	table[count].sym=sym;
	table[count].child=0;
	count++;
}

//! \param code  
//! \param sym     
//! \result   (0 -   )
int LZW::Find(int code,BYTE sym)
{
	code=table[code].child;
	while(code) {
		if(table[code].sym==sym) break;
		code=table[code].next;
	}
	return code;
}

void LZW_Encoder::WriteCode(FILE *out,int code)
{
	for(int n=0; n<bits_count; n++) {
		if(!byte_shift) fputc(byte,out),byte_shift=1,byte=0;
		if(code&1) byte|=byte_shift;
		byte_shift<<=1;
		code>>=1;
	}
}

void LZW_Encoder::_Encode(FILE *in,FILE *out)
{
	LZW_Encoder obj;
	obj.Pack(in,out);
}


//! \param in   
//! \param out  
void LZW_Encoder::Pack(FILE *in,FILE *out)
{
	int code,sym,index;
	code=0;
	Init(); StartWrite();
	for(;;) {
		sym=fgetc(in); if(sym==EOF) break;
		if(code) index=Find(code,sym); else index=sym+1;
		if(index) code=index;
		else {
			WriteCode(out,code);
			if(count>=CAPABLITY) Init();
			Update(code,sym);
			if(count==(1<<bits_count)) bits_count++;
			code=sym+1;
		}
	}
	WriteCode(out,code);
	WriteCode(out,LZ_END);
	StopWrite(out);
}

LZW_Decoder::LZW_Decoder()
{
	stack=0; stack_size=0;
}

LZW_Decoder::~LZW_Decoder()
{
	free(stack);
}



//! \param out   
//! \param code     
void LZW_Decoder::WriteSym(FILE *out,int code)
{
	int size=0;
	for(;;) {
		if(!code) break;
		if(size>=stack_size) {
			stack_size+=((stack_size<<1)+9)&~7;
			stack=(char*)realloc(stack,stack_size);
		}
		stack[size++]=table[code].sym;
		code=table[code].prefix;
	}
	while(size) size--,fputc(stack[size],out);
}


//! \param in   
//! \result  
int LZW_Decoder::ReadCode(FILE *in)
{
	int code=0;
	for(int n=0; n<bits_count; n++) {
		if(!byte_shift) byte=fgetc(in),byte_shift=1;
		if(byte&byte_shift) code|=1<<n;
		byte_shift<<=1;
	}
	return code;
}


void LZW_Decoder::_Decode(FILE *in,FILE *out)
{
	LZW_Decoder obj;
	obj.Unpack(in,out);
}


//! \param in   
//! \param out  
void LZW_Decoder::Unpack(FILE *in,FILE *out)
{
	int n,old_code,sym,new_code;
	old_code=0;
	Init(); StartRead();
	for(;;) {
		if(count>=CAPABLITY) Init();
		new_code=ReadCode(in);
		if(!new_code) break;
		if(new_code>=count) WriteSym(out,old_code),WriteSym(out,sym+1),n=old_code;
		else WriteSym(out,new_code),n=new_code;
		while(n) sym=table[n].sym,n=table[n].prefix;
		if(old_code) {
			Update(old_code,sym);
			if(count>=(1<<bits_count)-1) bits_count++;
		}
		old_code=new_code;
	}
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	FILE *input =fopen(list[2],"rb");
	FILE *output=fopen(list[3],"wb");
	if(!input || !output) {
		printf("IO error\n\a");
		return 3;
	}
	switch(list[1][1]) {
	case 'e': LZW_Encoder::_Encode(input,output); break;	// 
	case 'd': LZW_Decoder::_Decode(input,output); break;	// 
	}
	fclose(input);
	fclose(output);
	return 0;
}


